#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from torchvision.datasets.folder import default_loader


class ListDataset:
    """Dataset that loads data using a list of items, a corresponding loader,
    and a list of metadata. The default loader is an image file loader so this
    dataset can be used directly with a list of image files.
    You can use it without metadata if you set metadata parameter to None
    """

    def __init__(self, files, metadata, loader=default_loader):
        """
        metadata (List[Dict[Type]] or List[Type], Optional):
            metadata to be added to each sample.
            The Type can be anything that pytorch default_collate can handle.
            If Type is tensor, make sure that the tensors are of same dimension.
        """
        if metadata is not None:
            assert isinstance(metadata, list), "metadata should be a list"
            assert len(files) == len(metadata)
            assert len(files) > 0, "Empty ListDataset is not allowed"
            if not isinstance(metadata[0], dict):
                metadata = [{"target": target} for target in metadata]
        self.files = files
        self.metadata = metadata
        self.loader = loader

    def __getitem__(self, idx):
        assert idx >= 0 and idx < len(self)
        img = self.loader(self.files[idx])
        item = {"input": img}
        if self.metadata is not None:
            item.update(self.metadata[idx])

        return item

    def __len__(self):
        return len(self.files)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.losses import (
    ClassyLoss,
    MultiOutputSumLoss,
    build_loss,
    register_loss,
)


@register_loss("mock_1")
class MockLoss1(ClassyLoss):
    def forward(self, pred, target):
        return torch.tensor(1.0)

    @classmethod
    def from_config(cls, config):
        return cls()


class TestMultiOutputSumLoss(unittest.TestCase):
    def test_multi_output_sum_loss(self):
        config = {"name": "multi_output_sum_loss", "loss": {"name": "mock_1"}}
        crit = build_loss(config)
        self.assertTrue(isinstance(crit, MultiOutputSumLoss))

        # test with a single output
        output = torch.tensor([1.0, 2.3])
        target = torch.tensor(1.0)
        self.assertAlmostEqual(crit(output, target).item(), 1.0)

        # test with a list of outputs
        output = [torch.tensor([1.2, 3.2])] * 5
        target = torch.tensor(2.3)
        self.assertAlmostEqual(crit(output, target).item(), 5.0)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os
from typing import Optional

import torch
from classy_vision.generic.distributed_util import (
    get_rank,
    get_world_size,
    set_cpu_device,
    set_cuda_device_index,
)

from .classy_trainer import ClassyTrainer


def _init_env_vars():
    """Function sets up default environment variables for distributed training.
    """
    if "WORLD_SIZE" not in os.environ or "RANK" not in os.environ:
        os.environ["WORLD_SIZE"] = "1"
        os.environ["RANK"] = "0"
        os.environ["LOCAL_RANK"] = "0"

    if "MASTER_ADDR" not in os.environ or "MASTER_PORT" not in os.environ:
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "29500"


def _init_distributed(use_gpu: bool):
    """Function perform distributed setup for DDP.

    Requires the script to be started with torch.distributed.launch
    script and uses environment variables for node finding.

    Args:
        use_gpu: If true, use distributed GPU training, else use CPU
    """
    distributed_world_size = int(os.environ["WORLD_SIZE"])
    distributed_rank = int(os.environ["RANK"])
    backend = "nccl" if use_gpu else "gloo"
    torch.distributed.init_process_group(
        backend=backend,
        init_method="env://",
        world_size=distributed_world_size,
        rank=distributed_rank,
    )


class DistributedTrainer(ClassyTrainer):
    """Distributed trainer for using multiple training processes
    """

    def __init__(
        self,
        use_gpu: Optional[bool] = None,
        num_dataloader_workers: int = 0,
        dataloader_mp_context: Optional[str] = None,
    ):
        """Constructor for DistributedTrainer.

        Args:
            use_gpu: If true, then use GPU 0 for training.
                If None, then check if we have GPUs available, if we do
                then use GPU for training.
            num_dataloader_workers: Number of CPU processes doing dataloading
                per GPU. If 0, then dataloading is done on main thread.
            dataloader_mp_context: Determines how to launch
                new processes for dataloading. Must be one of "fork", "forkserver",
                "spawn". If None, process launching is inherited from parent.
        """
        super().__init__(
            use_gpu=use_gpu,
            num_dataloader_workers=num_dataloader_workers,
            dataloader_mp_context=dataloader_mp_context,
        )
        _init_env_vars()
        _init_distributed(self.use_gpu)
        logging.info(
            f"Done setting up distributed process_group with rank {get_rank()}"
            + f", world_size {get_world_size()}"
        )
        local_rank = int(os.environ["LOCAL_RANK"])
        if self.use_gpu:
            logging.info("Using GPU, CUDA device index: {}".format(local_rank))
            set_cuda_device_index(local_rank)
        else:
            logging.info("Using CPU")
            set_cpu_device()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.losses import build_loss


class CriterionsTest(unittest.TestCase):
    """
    Test that build_transform is able to build torch losses correctly.
    """

    def _test_loss(self, config, output, target, expected_loss):
        # test that we are able to build losses from torch.nn.modules.loss
        # and that they work correctly

        crit = build_loss(config)

        # test that the weights are set correctly
        self.assertAlmostEqual(crit.weight.numpy().tolist(), [1.0, 1.0])

        # test that the loss is computed correctly
        self.assertAlmostEqual(crit(output, target).item(), expected_loss)

        # verify ignore index works
        if "ignore_index" in config:
            self.assertAlmostEqual(crit(output, torch.tensor([-1])).item(), 0.0)

    def test_cross_entropy_loss(self):
        """
        Test CrossEntropyLoss
        """
        config = {
            "name": "CrossEntropyLoss",
            "weight": [1.0, 1.0],
            "ignore_index": -1,
            "reduction": "mean",
        }
        output = torch.tensor([[9.0, 1.0]])
        target = torch.tensor([1])
        expected_loss = 8.000335693359375
        self._test_loss(config, output, target, expected_loss)

    def test_bce_with_logits_loss(self):
        """
        Test BCEWithLogitsLoss
        """
        config = {
            "name": "BCEWithLogitsLoss",
            "weight": [1.0, 1.0],
            "reduction": "mean",
        }
        output = torch.tensor([0.999, 0.999])
        target = torch.tensor([1.0, 1.0])
        expected_loss = 0.313530727260701
        self._test_loss(config, output, target, expected_loss)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import os
import re
import sys

from setuptools import find_packages, setup


if __name__ == "__main__":
    if sys.version_info < (3, 6):
        sys.exit("Sorry, Python >=3.6 is required for Classy Vision.")

    # get version string from module
    with open(
        os.path.join(os.path.dirname(__file__), "classy_vision/__init__.py"), "r"
    ) as f:
        version = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M).group(
            1
        )
        print("-- Building version " + version)

    with open("README.md", encoding="utf8") as f:
        readme = f.read()

    with open("requirements.txt") as f:
        reqs = f.read()

    setup(
        name="classy_vision",
        version=version,
        description="An end-to-end PyTorch framework for image and video classification.",
        long_description_content_type="text/markdown",
        long_description=readme,
        url="https://classyvision.ai",
        project_urls={
            "Documentation": "https://classyvision.ai",
            "Source": "https://github.com/facebookresearch/ClassyVision",
        },
        license="MIT License",
        python_requires=">=3.6",
        packages=find_packages(exclude=("tests",)),
        install_requires=reqs.strip().split("\n"),
        extras_require={
            "dev": [
                "GitPython",
                "black==19.3b0",
                "sphinx",
                "isort",
                "bs4",
                "nbconvert",
                "pre-commit",
            ]
        },
        package_data={"classy_vision": ["configs/*.json", "templates"]},
        data_files=[("classy_vision", ["classy_train.py"])],
        include_package_data=True,
        test_suite="test.suites.unittests",
        scripts=["bin/classy-project"],
        keywords=["deep learning", "pytorch", "AI"],
        classifiers=[
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Image Recognition",
        ],
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""MLP model."""

from typing import Any, Dict

import torch.nn as nn

from . import register_model
from .classy_model import ClassyModel


@register_model("mlp")
class MLP(ClassyModel):
    """MLP model using ReLU. Useful for testing on CPUs."""

    def __init__(
        self,
        input_dim,
        output_dim,
        hidden_dims,
        dropout,
        first_dropout,
        use_batchnorm,
        first_batchnorm,
    ):
        super().__init__()

        layers = []
        # If first_batchnorm is set, must be using batchnorm
        assert not first_batchnorm or use_batchnorm

        self._num_inputs = input_dim
        self._num_classes = output_dim
        self._model_depth = len(hidden_dims) + 1

        if dropout > 0 and first_dropout:
            layers.append(nn.Dropout(p=dropout))

        if use_batchnorm and first_batchnorm:
            layers.append(nn.BatchNorm1d(input_dim))

        for dim in hidden_dims:
            layers.append(nn.Linear(input_dim, dim))
            if use_batchnorm:
                layers.append(nn.BatchNorm1d(dim))
            if dropout > 0:
                layers.append(nn.Dropout(p=dropout))
            layers.append(nn.ReLU(inplace=True))
            input_dim = dim

        layers.append(nn.Linear(input_dim, output_dim))
        self.mlp = nn.Sequential(*layers)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "MLP":
        """Instantiates a MLP from a configuration.

        Args:
            config: A configuration for a MLP.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A MLP instance.
        """
        assert (key in config for key in ["input_dim", "output_dim", "hidden_dims"])

        output_dim = config["output_dim"]
        return cls(
            input_dim=config["input_dim"],
            output_dim=output_dim,
            hidden_dims=config["hidden_dims"],
            dropout=config.get("dropout", 0),
            first_dropout=config.get("first_dropout", False),
            use_batchnorm=config.get("use_batchnorm", False),
            first_batchnorm=config.get("first_batchnorm", False),
        )

    def forward(self, x):
        batchsize_per_replica = x.shape[0]
        out = x.view(batchsize_per_replica, -1)
        out = self.mlp(out)
        return out

    @property
    def input_shape(self):
        return (self._num_inputs,)

    @property
    def output_shape(self):
        return (1, self._num_classes)

    @property
    def model_depth(self):
        return self._model_depth



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum, auto
from typing import Any, Dict, Sequence

from . import (
    ClassyParamScheduler,
    UpdateInterval,
    build_param_scheduler,
    register_param_scheduler,
)


class IntervalScaling(Enum):
    RESCALED = auto()
    FIXED = auto()


@register_param_scheduler("composite")
class CompositeParamScheduler(ClassyParamScheduler):
    """
    Composite parameter scheduler composed of intermediate schedulers.
    Takes a list of schedulers and a list of lengths corresponding to
    percentage of training each scheduler should run for. Schedulers
    are run in order. All values in lengths should sum to 1.0.

    Each scheduler also has a corresponding interval scale. If interval
    scale is 'fixed', the intermediate scheduler will be run without any rescaling
    of the time. If interval scale is 'rescaled', intermediate scheduler is
    run such that each scheduler will start and end at the same values as it
    would if it were the only scheduler. Default is 'rescaled' for all schedulers.

    Example:

        .. code-block:: python

              update_interval = "step"
              schedulers = [
                {"name": "constant", "value": 0.42},
                {"name": "cosine_decay", "start_lr": 0.42, "end_lr": 0.0001}
              ]
              interval_scaling = ['rescaled', 'rescaled'],
              lengths =  [0.3, 0.7]

    The parameter value will be 0.42 for the first [0%, 30%) of steps,
    and then will cosine decay from 0.42 to 0.0001 for [30%, 100%) of
    training.
    """

    def __init__(
        self,
        schedulers: Sequence[ClassyParamScheduler],
        lengths: Sequence[float],
        update_interval: UpdateInterval,
        interval_scaling: Sequence[IntervalScaling],
    ):
        super().__init__()
        self.update_interval = update_interval
        self._lengths = lengths
        self._schedulers = schedulers
        self._interval_scaling = interval_scaling

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "CompositeParamScheduler":
        """Instantiates a CompositeParamScheduler from a configuration.

        Args:
            config: A configuration for a CompositeParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A CompositeParamScheduler instance.
        """
        assert (
            "schedulers" in config and "lengths" in config
        ), "Composite scheduler needs both a list of schedulers and lengths"
        assert len(config["schedulers"]) == len(
            config["lengths"]
        ), "Schedulers and lengths must be same length"
        assert (
            len(config["schedulers"]) > 0
        ), "There must be at least one scheduler in the composite scheduler"
        assert (
            abs(sum(config["lengths"]) - 1.0) < 1e-3
        ), "The sum of all values in lengths must be 1"
        if sum(config["lengths"]) != 1.0:
            config["lengths"][-1] = 1.0 - sum(config["lengths"][:-1])
        update_interval = UpdateInterval.STEP
        if "update_interval" in config:
            assert config["update_interval"] in {
                "step",
                "epoch",
            }, "Choices for update interval are 'step' or 'epoch'"
            update_interval = UpdateInterval[config["update_interval"].upper()]
        interval_scaling = []
        if "interval_scaling" in config:
            assert len(config["schedulers"]) == len(
                config["interval_scaling"]
            ), "Schedulers and interval scaling must be the same length"
            for interval_scale in config["interval_scaling"]:
                assert interval_scale in {
                    "fixed",
                    "rescaled",
                }, "Choices for interval scaling are 'fixed' or 'rescaled'"
                interval_scaling.append(IntervalScaling[interval_scale.upper()])
        else:
            interval_scaling = [IntervalScaling.RESCALED] * len(config["schedulers"])
        if "num_epochs" in config:  # Propagate value to intermediate schedulers
            config["schedulers"] = [
                dict(schedule, **{"num_epochs": config["num_epochs"]})
                for schedule in config["schedulers"]
            ]
        return cls(
            schedulers=[
                build_param_scheduler(scheduler) for scheduler in config["schedulers"]
            ],
            lengths=config["lengths"],
            update_interval=update_interval,
            interval_scaling=interval_scaling,
        )

    def __call__(self, where: float):
        # Find scheduler corresponding to where
        i = 0
        running_total = self._lengths[i]
        while (where + self.WHERE_EPSILON) > running_total and i < len(
            self._schedulers
        ) - 1:
            i += 1
            running_total += self._lengths[i]
        scheduler = self._schedulers[i]
        scheduler_where = where
        interval_scale = self._interval_scaling[i]
        if interval_scale == IntervalScaling.RESCALED:
            # Calculate corresponding where % for scheduler
            scheduler_start = running_total - self._lengths[i]
            scheduler_where = (where - scheduler_start) / self._lengths[i]
        return scheduler(scheduler_where)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch

from . import register_transform
from .classy_transform import ClassyTransform


_IMAGENET_EIGEN_VAL = [0.2175, 0.0188, 0.0045]
_IMAGENET_EIGEN_VEC = [
    [-144.7125, 183.396, 102.2295],
    [-148.104, -1.1475, -207.57],
    [-148.818, -177.174, 107.1765],
]

_DEFAULT_COLOR_LIGHTING_STD = 0.1


@register_transform("lighting")
class LightingTransform(ClassyTransform):
    """
    Lighting noise(AlexNet - style PCA - based noise).
    This trick was originally used in `AlexNet paper
    <https://papers.nips.cc/paper/4824-imagenet-classification
    -with-deep-convolutional-neural-networks.pdf>`_

    The eigen values and eigen vectors, are taken from caffe2 `ImageInputOp.h
    <https://github.com/pytorch/pytorch/blob/master/caffe2/image/
    image_input_op.h#L265>`_.
    """

    def __init__(
        self,
        alphastd=_DEFAULT_COLOR_LIGHTING_STD,
        eigval=_IMAGENET_EIGEN_VAL,
        eigvec=_IMAGENET_EIGEN_VEC,
    ):
        self.alphastd = alphastd
        # Divide by 255 as the Lighting operation is expected to be applied
        # on `img` pixels ranging between [0.0, 1.0]
        self.eigval = torch.tensor(eigval) / 255.0
        self.eigvec = torch.tensor(eigvec) / 255.0

    def __call__(self, img):
        """
            img: (C x H x W) Tensor with values in range [0.0, 1.0]
        """
        assert (
            img.min() >= 0.0 and img.max() <= 1.0
        ), "Image should be normalized by 255 and be in range [0.0, 1.0]"
        if self.alphastd == 0:
            return img

        alpha = img.new().resize_(3).normal_(0, self.alphastd)
        rgb = (
            self.eigvec.type_as(img)
            .clone()
            .mul(alpha.view(1, 3).expand(3, 3))
            .mul(self.eigval.view(1, 3).expand(3, 3))
            .sum(1)
            .squeeze()
        )

        return img.add(rgb.view(3, 1, 1).expand_as(img))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Tuple

import torch


class ClassyMeter:
    """
    Base class to measure various metrics during training and testing phases.

    This can include meters like  Accuracy, Precision and Recall, etc.
    """

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyMeter":
        """Instantiates a ClassyMeter using a configuration.

        Args:
            config: A configuration for a ClassyMeter.

        Returns:
            A ClassyMeter instance.
        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """The name of the meter."""
        raise NotImplementedError

    @property
    def value(self) -> Any:
        """
        Value of meter based on local state, can be any python object.

        Note:
            If there are multiple training processes then this
            represents the local state of the meter. If :func:`sync_state` is
            implemented, then value will return the global state since the
            last sync PLUS any local unsynced updates that have occurred
            in the local process.
        """
        raise NotImplementedError

    def sync_state(self) -> None:
        """
        Syncs state with all other meters in distributed training.

        If not provided by child class this does nothing by default
        and meter only provides the local process stats. If
        implemented then the meter provides the global stats at last
        sync + any local updates since the last sync.

        Warning:
            Calls to sync_state could involve communications via
            :mod:`torch.distributed` which can result in a loss of performance or
            deadlocks if not coordinated among threads.
        """
        pass

    def reset(self):
        """
        Resets any internal meter state.

        Should normally be called at the end of a phase.
        """
        raise NotImplementedError

    def update(
        self, model_output: torch.Tensor, target: torch.Tensor, **kwargs
    ) -> None:
        """
        Updates any internal state of meter.

        Should be called after each batch processing of each phase.

        Args:
            model_output: Output of a :class:`ClassyModel`.
            target: Target provided by a dataloader from :class:`ClassyDataset`.
        """
        raise NotImplementedError

    def validate(self, model_output_shape: Tuple, target_shape: Tuple) -> None:
        """
        Validate the meter.

        Checks if the meter can be calculated on the given ``model_output_shape``
        and ``target_shape``.
        """
        raise NotImplementedError

    def get_classy_state(self) -> Dict[str, Any]:
        """Get the state of the ClassyMeter.

        The returned state is used for checkpointing.

        Returns:
            A state dictionary containing the state of the meter.
        """
        raise NotImplementedError

    def set_classy_state(self, state: Dict[str, Any]) -> None:
        """Set the state of the ClassyMeter.

        Args:
            state_dict: The state dictionary. Must be the output of a call to
                :func:`get_classy_state`.

        This is used to load the state of the meter from a checkpoint.
        """
        raise NotImplementedError



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.heads import ClassyHead
from classy_vision.models import ClassyModel


class TestClassyBlock(unittest.TestCase):
    class DummyTestHead(ClassyHead):
        def __init__(self):
            super().__init__("head_id")
            self.layer = torch.nn.Linear(2, 2)

        def forward(self, x):
            return self.layer(x)

    class DummyTestModel(ClassyModel):
        def __init__(self):
            super().__init__()
            self.layer1 = self.build_attachable_block(
                "dummy_block", torch.nn.Linear(2, 2)
            )
            self.layer2 = self.build_attachable_block(
                "dummy_block2", torch.nn.Linear(2, 2)
            )

        def forward(self, x):
            out = self.layer1(x)
            return self.layer2(out)

    def test_head_execution(self):
        model = self.DummyTestModel()
        head = self.DummyTestHead()
        model.set_heads({"dummy_block2": {head.unique_id: head}})
        input = torch.randn(1, 2)
        output = model(input)
        head_output = model.execute_heads()
        self.assertTrue(torch.allclose(head(output), head_output["head_id"]))

    def test_duplicated_head_ids(self):
        model = self.DummyTestModel()
        head1 = self.DummyTestHead()
        head2 = self.DummyTestHead()
        heads = {
            "dummy_block": {head1.unique_id: head1},
            "dummy_block2": {head2.unique_id: head2},
        }
        with self.assertRaises(ValueError):
            model.set_heads(heads)

        head2.unique_id = "head_id2"
        model.set_heads(heads)

    def test_set_heads(self):
        model = self.DummyTestModel()
        head = self.DummyTestHead()
        self.assertEqual(
            len(model.get_heads()), 0, "heads should be empty before set_heads"
        )
        model.set_heads({"dummy_block2": {head.unique_id: head}})
        input = torch.randn(1, 2)
        model(input)
        head_outputs = model.execute_heads()
        self.assertEqual(len(head_outputs), 1, "should have output for one head")

        # remove all heads
        model.set_heads({})
        self.assertEqual(len(model.get_heads()), 0, "heads should be empty")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from classy_vision.generic.util import is_pos_int
from classy_vision.meters.accuracy_meter import AccuracyMeter

from . import register_meter
from .video_meter import VideoMeter


@register_meter("video_accuracy")
class VideoAccuracyMeter(VideoMeter):
    """Meter to calculate top-k video-level accuracy for single/multi label
       video classification task.

    Video-level accuarcy is computed by averaging clip-level predictions and
    compare the reslt with video-level groundtruth label.
    """

    def __init__(self, topk, clips_per_video_train, clips_per_video_test):
        """
        Args:
            topk: list of int `k` values.
            clips_per_video_train: No. of clips sampled per video at train time
            clips_per_video_test: No. of clips sampled per video at test time
        """
        super().__init__(clips_per_video_train, clips_per_video_test)
        assert isinstance(topk, list), "topk must be a list"
        assert len(topk) > 0, "topk list should have at least one element"
        assert [is_pos_int(x) for x in topk], "each value in topk must be >= 1"

        self._accuracy_meter = AccuracyMeter(topk)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "VideoAccuracyMeter":
        """Instantiates a VideoAccuracyMeter from a configuration.

        Args:
            config: A configuration for a VideoAccuracyMeter.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A VideoAccuracyMeter instance.
        """
        return cls(
            topk=config["topk"],
            clips_per_video_train=config.get("clips_per_video_train", 1),
            clips_per_video_test=config["clips_per_video_test"],
        )

    @property
    def name(self):
        return "video_accuracy"

    @property
    def meter(self):
        return self._accuracy_meter



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Tuple

import torch.optim

from . import ClassyOptimizer, register_optimizer


@register_optimizer("adam")
class Adam(ClassyOptimizer):
    def __init__(
        self,
        lr: float = 0.1,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 0.0,
        amsgrad: bool = False,
    ) -> None:
        super().__init__()

        self.parameters.lr = lr
        self.parameters.betas = betas
        self.parameters.eps = eps
        self.parameters.weight_decay = weight_decay
        self.parameters.amsgrad = amsgrad

    def init_pytorch_optimizer(self, model, **kwargs) -> None:
        super().init_pytorch_optimizer(model, **kwargs)
        self.optimizer = torch.optim.Adam(
            self.param_groups_override,
            lr=self.parameters.lr,
            betas=self.parameters.betas,
            eps=self.parameters.eps,
            weight_decay=self.parameters.weight_decay,
            amsgrad=self.parameters.amsgrad,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Adam":
        """Instantiates a Adam from a configuration.

        Args:
            config: A configuration for a Adam.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A Adam instance.
        """
        # Default params
        config.setdefault("lr", 0.1)
        config.setdefault("betas", [0.9, 0.999])
        config.setdefault("eps", 1e-8)
        config.setdefault("weight_decay", 0.0)
        config.setdefault("amsgrad", False)

        # Check if betas is a list and convert it to a tuple
        # since a JSON config can only have lists
        if type(config["betas"]) == list:
            config["betas"] = tuple(config["betas"])

        assert (
            type(config["betas"]) == tuple
            and len(config["betas"]) == 2
            and type(config["betas"][0]) == float
            and type(config["betas"][1]) == float
            and config["betas"][0] >= 0.0
            and config["betas"][0] < 1.0
            and config["betas"][1] >= 0.0
            and config["betas"][1] < 1.0
        ), "Config must contain a tuple 'betas' in [0, 1) for Adam optimizer"

        return cls(
            lr=config["lr"],
            betas=config["betas"],
            eps=config["eps"],
            weight_decay=config["weight_decay"],
            amsgrad=config["amsgrad"],
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import shutil
import tempfile
import unittest

import PIL
from classy_vision.dataset import build_dataset
from classy_vision.dataset.classy_imagenet import ImageNetDataset
from torchvision import transforms


class TestImageNet(unittest.TestCase):
    def get_test_image_dataset(self, num_samples):
        config = {
            "name": "synthetic_image",
            "crop_size": 224,
            "num_channels": 3,
            "seed": 0,
            "class_ratio": 0.5,
            "num_samples": num_samples,
            "batchsize_per_replica": 1,
            "use_shuffle": False,
            "transforms": [
                {
                    "name": "apply_transform_to_key",
                    "transforms": [{"name": "ToTensor"}],
                    "key": "input",
                }
            ],
        }
        dataset = build_dataset(config)
        return dataset

    def setUp(self):
        # create a base directory to write image files to
        self.base_dir = tempfile.mkdtemp()
        for split in ["train", "val"]:
            os.makedirs(f"{self.base_dir}/{split}/0")
            os.makedirs(f"{self.base_dir}/{split}/1")

    def tearDown(self):
        # delete all the temporary data created
        shutil.rmtree(self.base_dir)

    def test_imagenet_retrieve_sample(self):
        num_samples = 10
        for split in ["train", "val"]:
            dataloader = self.get_test_image_dataset(num_samples).iterator()
            for i, sample in enumerate(dataloader):
                input = sample["input"]
                target = sample["target"]
                image = transforms.ToPILImage()(input.squeeze())
                path = f"{self.base_dir}/{split}/{target.item()}/{i}.png"
                # save the image in a lossless format (png)
                image.save(path)

            dataset = ImageNetDataset(
                split="train",
                batchsize_per_replica=1,
                shuffle=True,
                transform=None,
                num_samples=None,
                root=self.base_dir,
            )
            self.assertEqual(len(dataset), num_samples)
            img, target = dataset[0]
            self.assertTrue(isinstance(img, PIL.Image.Image))
            self.assertTrue(isinstance(target, int))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging

import torch
from classy_vision.meters import ClassyMeter


class VideoMeter(ClassyMeter):
    """An abstraction of meter for evaluating video models.

    Video-level metric is computed by averaging clip-level predictions and
    compare the result with video-level groundtruth label.

    This meter abstraction can wrap conventional classy meters by passing
    averaged clip-level predictions to the meter needed for video level metrics.
    """

    def __init__(self, clips_per_video_train, clips_per_video_test):
        """Constructor of VideoMeter class.

        Args:
            clips_per_video_train: No. of clips sampled per video at train time
            clips_per_video_test: No. of clips sampled per video at test time
        """

        self._clips_per_video_train = clips_per_video_train
        self._clips_per_video_test = clips_per_video_test

    @property
    def value(self):
        return self.meter.value

    def sync_state(self):
        self.meter.sync_state()

    @property
    def meter(self) -> "ClassyMeter":
        """Every video meter should implement to have its own internal meter.

        It consumes the video level predictions and ground truth label, and compute
        the actual metrics.

        Returns:
            An instance of ClassyMeter.
        """
        raise NotImplementedError

    def get_classy_state(self):
        """Contains the states of the meter.
        """
        state = {}
        state["meter_state"] = self.meter.get_classy_state()
        state["name"] = self.name
        state["clips_per_video_train"] = self._clips_per_video_train
        state["clips_per_video_test"] = self._clips_per_video_test
        return state

    def set_classy_state(self, state):
        assert (
            self.name == state["name"]
        ), "State name {state_name} does not match meter name {obj_name}".format(
            state_name=state["name"], obj_name=self.name
        )
        assert (
            self._clips_per_video_train == state["clips_per_video_train"]
        ), "incompatible clips_per_video_train for video accuracy"
        assert (
            self._clips_per_video_test == state["clips_per_video_test"]
        ), "incompatible clips_per_video_test for video accuracy"
        # Restore the state -- correct_predictions and sample_count.
        self.reset()
        self.meter.set_classy_state(state["meter_state"])

    def __repr__(self):
        return repr({"name": self.name, "value": self.value})

    def update(self, model_output, target, is_train, **kwargs):
        """Updates any internal state of meter with new model output and target.

        Args:
            model_output: tensor of shape (B * clips_per_video, C) where each value is
                          either logit or class probability.
            target:       tensor of shape (B * clips_per_video).
            is_train     if True, it is training stage when meter is updated

            Note: For binary classification, C=2.
        """
        num_clips = len(model_output)
        clips_per_video = (
            self._clips_per_video_train if is_train else self._clips_per_video_test
        )

        if not num_clips % clips_per_video == 0:
            logging.info(
                "Skip meter update. Because for video model testing, batch size "
                "is expected to be a multplier of No. of clips per video. "
                "num_clips: %d, clips_per_video: %d" % (num_clips, clips_per_video)
            )
            return

        num_videos = num_clips // clips_per_video
        for i in range(num_videos):
            clip_labels = target[i * clips_per_video : (i + 1) * clips_per_video]
            if clip_labels.ndim == 1:
                # single label
                assert (
                    len(torch.unique(clip_labels)) == 1
                ), "all clips from the same video should have same label"
            elif clip_labels.ndim == 2:
                # multi-hot label
                for j in range(1, clip_labels.shape[0]):
                    assert torch.equal(
                        clip_labels[0], clip_labels[j]
                    ), "all clips from the same video should have the same labels"
            else:
                raise ValueError(
                    "dimension of clip label matrix should be either 1 or 2"
                )

        video_model_output = torch.mean(
            torch.reshape(model_output, (num_videos, clips_per_video, -1)), 1
        )
        video_target = target[::clips_per_video]
        self.meter.update(video_model_output, video_target)

    def reset(self):
        self.meter.reset()

    def validate(self, model_output_shape, target_shape):
        self.meter.validate(model_output_shape, target_shape)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .classy_trainer import ClassyTrainer
from .distributed_trainer import DistributedTrainer
from .local_trainer import LocalTrainer


__all__ = ["ClassyTrainer", "DistributedTrainer", "LocalTrainer"]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from enum import Enum

import numpy as np
from PIL import Image

from ...generic.util import numpy_seed


class SampleType(Enum):
    DICT = "dict"
    TUPLE = "tuple"
    LIST = "list"


def _get_typed_sample(input, target, sample_type):
    if sample_type == SampleType.DICT:
        return {"input": input, "target": target}
    elif sample_type == SampleType.TUPLE:
        return (input, target)
    elif sample_type == SampleType.LIST:
        return [input, target]
    else:
        raise TypeError("Provided sample_type is not dict, list, tuple")


class RandomImageDataset:
    def __init__(
        self,
        crop_size,
        num_channels,
        num_classes,
        num_samples,
        seed,
        sample_type=SampleType.DICT,
    ):
        self.crop_size = crop_size
        self.num_channels = num_channels
        self.num_classes = num_classes
        self.num_samples = num_samples
        self.seed = seed
        self.sample_type = sample_type

    def __getitem__(self, idx):
        with numpy_seed(self.seed + idx):
            input = Image.fromarray(
                (
                    np.random.standard_normal(
                        [self.crop_size, self.crop_size, self.num_channels]
                    )
                    * 255
                ).astype(np.uint8)
            )
            target = np.random.randint(self.num_classes)
            return _get_typed_sample(input, target, self.sample_type)

    def __len__(self):
        return self.num_samples


class RandomImageBinaryClassDataset:
    def __init__(
        self, crop_size, class_ratio, num_samples, seed, sample_type=SampleType.DICT
    ):
        self.crop_size = crop_size
        # User Defined Class Imbalace Ratio
        self.class_ratio = class_ratio
        self.num_samples = num_samples
        self.seed = seed
        self.sample_type = sample_type

    def __getitem__(self, idx):
        with numpy_seed(self.seed + idx):
            class_id = int(np.random.random() < self.class_ratio)
            image = np.zeros((self.crop_size, self.crop_size, 3))
            image[:, :, class_id] = np.random.random([self.crop_size, self.crop_size])
            image[:, :, 2] = np.random.random([self.crop_size, self.crop_size])
            input = Image.fromarray((image * 255).astype(np.uint8))
            target = class_id
            return _get_typed_sample(input, target, self.sample_type)

    def __len__(self):
        return self.num_samples



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Optional, Tuple

import torch.nn as nn

from .classy_model import ClassyModel


class ClassyModelWrapper(ClassyModel):
    """
    Class which wraps an `nn.Module <https://pytorch.org/docs/stable/
    nn.html#torch.nn.Module>`_ within a ClassyModel.

    The only required argument is the model, the additional args are needed
    to get some additional capabilities from Classy Vision to work.
    """

    def __init__(
        self,
        model: nn.Module,
        input_shape: Optional[Tuple] = None,
        output_shape: Optional[Tuple] = None,
        model_depth: Optional[int] = None,
    ):
        super().__init__()
        self.model = model
        self._input_shape = input_shape
        self._output_shape = output_shape
        self._model_depth = model_depth

    def forward(self, x):
        return self.model(x)

    def extract_features(self, x):
        if hasattr(self.model, "extract_features"):
            return self.model.extract_features(x)
        return super().extract_features(x)

    @property
    def input_shape(self):
        if self._input_shape is not None:
            return self._input_shape
        return super().input_shape

    @property
    def output_shape(self):
        if self._output_shape is not None:
            return self._output_shape
        return super().output_shape

    @property
    def model_depth(self):
        if self._model_depth is not None:
            return self._model_depth
        return super().model_depth



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
from classy_vision.generic.util import convert_to_one_hot


class TestUtils(unittest.TestCase):
    def test_single(self):
        targets = torch.tensor([[4]])
        one_hot_target = convert_to_one_hot(targets, 5)
        self.assertTrue(torch.allclose(one_hot_target, torch.tensor([[0, 0, 0, 0, 1]])))

    def test_two(self):
        targets = torch.tensor([[0], [1]])
        one_hot_target = convert_to_one_hot(targets, 3)
        self.assertTrue(
            torch.allclose(one_hot_target, torch.tensor([[1, 0, 0], [0, 1, 0]]))
        )
